#   FitPoints.py
#       created on: 11 January 2017
#
#           Author: Andreas Gadin

import numpy as np
import sys

# Class of static methods that are used to fit one set of perturbed data points from a .xyz-file to another set
#  unperturbed data points. Input is given as <python Fitpoints.py unperturbed_data.xyz perturbed_data.xyz>
class FitPoints:
    @staticmethod
    def read(filename):
        f = open(filename, 'r')
        lines = f.readlines()
        f.close()
        return lines

    @staticmethod
    def openFile(filename, lines):
        file_extension = '.xyz'
        if filename.endswith(file_extension):
            filename = filename[:-len(file_extension)] + '_fit.xyz'
        else:
            print('Must be a .xyz - file')
            sys.exit()
        f = open(filename, 'w')
        return f

    # Takes a rotation matrix and a translation and uses them to fit a line.
    # The line is inputted as a string from the .xyz format and is outputted in the same format.
    @staticmethod
    def fit(R, d, line):
        line_sp = line.split(' ')
        if line_sp[0] is 'C':
            p = FitPoints.line2point(line_sp)
            p = FitPoints.rotate(R, p)
            p = FitPoints.translate(d, p)
            line = 'C ' + str(p[0]) + ' ' + str(p[1]) + ' ' + str(p[2]) + '\n'
        elif line_sp[0] is 'O':
            p = FitPoints.line2point(line_sp)
            p = FitPoints.rotate(R, p)
            p = FitPoints.translate(d, p)
            line = 'O ' + str(p[0]) + ' ' + str(p[1]) + ' ' + str(p[2]) + '\n'
        return line

    # Returns a 1 x 3 array of floats converted from a 1 x 4 array of strings
    @staticmethod
    def line2point(line):
        p = np.array([float(line[1]), float(line[2]), float(line[3])])
        return p

    #
    @staticmethod
    def rotate(R, p):
        p_prim = np.dot(R, p)
        return p_prim

    @staticmethod
    def translate(d, p):
        p_prim = p + d
        return p_prim

    # Returns a 3 x N matrix converted from string to floats (N = number_of_lines_with_char_C_in_it)
    @staticmethod
    def lines2matrix(lines):
        x, y, z = [], [], []
        for line in lines:
            line = line.split(' ')
            if line[0] is 'C' or line[0] is 'O':
                x.append(float(line[1]))
                y.append(float(line[2]))
                z.append(float(line[3]))
        return [x, y, z]

    # Returns the mean of each row of a 3 x N matrix as a 3 z 1 array.
    @staticmethod
    def mean(mat):
        xm = np.mean(mat[0])
        ym = np.mean(mat[1])
        zm = np.mean(mat[2])
        return np.array([xm, ym, zm])

    # Returns a 3 x N matrix subtracted by corresponding input.
    @staticmethod
    def subtractMean(mat, xm, ym, zm):
        return np.array([mat[0] - xm, mat[1] - ym, mat[2] - zm])

    # Method is only run if this script is invoked by the user.
    # Import class to run FitPoints from another script.
    @staticmethod
    def main():
        map = FitPoints.read(sys.argv[1])
        map_to = FitPoints.read(sys.argv[2])
        A = FitPoints.lines2matrix(map)
        B = FitPoints.lines2matrix(map_to)
        xm = FitPoints.mean(A)
        ym = FitPoints.mean(B)
        A = FitPoints.subtractMean(A, xm[0], xm[1], xm[2])
        B = FitPoints.subtractMean(B, ym[0], ym[1], ym[2])
        C = np.dot(B, np.transpose(A))
        U, s, Vt = np.linalg.svd(C, full_matrices=True)
        detUVt = np.linalg.det(np.dot(U, Vt))
        print('det(UVt)', detUVt)
        R = np.dot(U, np.dot(np.diag([1, 1, detUVt]), Vt))
        d = ym - np.dot(R, xm)

        # Translates and rotates all data points of argv[1] to argv[2] and
        # writes to a new file name as the original file, but
        # with _fit.xyz added to it.
        f = FitPoints.openFile(sys.argv[1], map)
        for line in map:
            line = FitPoints.fit(R, d, line)
            f.write(line)
        f.close()


if __name__ == '__main__':
    FitPoints.main()